/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <poll.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>

#include <digitalSTROM/ds485-socket-server.h>
#include <digitalSTROM/ds485-socket-interface.h>
#include "ds485-client-interface.h"
#include "ds485-client.h"
#include "util.h"
#include "base/log.h"
#include "base/socket_client.h"
#include "base/socket_send.h"

static int sockwrite(ds485_client_t *cl, uint8_t* data, size_t dlen);
static int socket_service_dispatch(ds485_client_t *cl, uint8_t* data, uint16_t pktlen);
static int client_connect(ds485_client_t *cl);
static void* socket_thread(void* arg);

static bool v1_link_state(bus_state_t _state)
{
    switch(_state) {
    case DS485_STATE_ISOLATED:
    case DS485_STATE_JOIN:
    case DS485_STATE_CONNECTED:
    case DS485_STATE_DISCONNECTED:
        return false;
    case DS485_STATE_ACTIVE:
        return true;
    }

    report(LOG_NOTICE, "unknown bus state: %d\n", _state );
    return true;
}

int ds485_socket_client_open(struct _ds485_client_s *cl)
{
    /* Start communication thread */
    int err;
    err = pthread_create(&cl->connThread, NULL, socket_thread, cl);
    if (err) {
        report(LOG_ERR, "pthread_create: connection thread %s\n",
               strerror(err));
        return ERROR_THREAD_START_FAILED;
    }
    return 0;
}

int ds485_socket_client_req_bus_state(ds485_client_t *cl)
{
    /* reply will be returned through callback */
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_QUERY_BUSSTATE;
    cmdCall.length = DS485N_TLV_HDR_LEN;

    return sockwrite(cl, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_get_dsuid(ds485_client_t *cl)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_QUERY_DSUID;
    cmdCall.length = DS485N_TLV_HDR_LEN;

    return sockwrite(cl, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_restart_bus(ds485_client_t *cl)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_RESTART_BUS;
    cmdCall.length = DS485N_TLV_HDR_LEN;

    return sockwrite(cl, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_get_bus_devices(ds485_client_t *cl, dsuid_t
                                        *device_list, int device_list_size)
{
    (void)device_list;
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_QUERY_DEVICES;
    cmdCall.length = DS485N_TLV_HDR_LEN + 1;
    cmdCall.data[0] = device_list_size;

    return sockwrite(cl, (uint8_t *) &cmdCall, cmdCall.length);
}

/**
 * Send dM-API command
 */
int ds485_socket_client_send_command(struct _ds485_client_s *cl,
                                     const ds485_container_t *request)
{
    dS485CmdApidata cmdCall;
    assert(request->length <= DS485_MAX_PAYLOAD_SIZE);

    memset(&cmdCall, 0, sizeof(dS485CmdApidata));
    cmdCall.apiData.destinationId = request->destinationId;
    cmdCall.apiData.sourceId = request->sourceId;
    cmdCall.apiData.containerType = request->containerType;
    cmdCall.apiData.length = request->length;
    cmdCall.apiData.transactionId = request->transactionId;
    memcpy(&cmdCall.apiData.data, request->data, cmdCall.apiData.length);

    cmdCall.length = DS485N_TLV_HDR_LEN + DS485N_CONT_LEN(&cmdCall.apiData);
    cmdCall.command = DS485D_CMD_TX_PACKET;

    return sockwrite(cl, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_send_long_command(struct _ds485_client_s *cl,
                                          const ds485n_packet_t *request)
{
    dS485CmdLongApidata longCmdCall;
    assert(request->length <= DS485N_PACKET_MAX_PAYLOAD_SIZE);

    memset(&longCmdCall, 0, sizeof(dS485CmdLongApidata));
    longCmdCall.apiData.destinationId = request->destinationId;
    longCmdCall.apiData.sourceId = request->sourceId;
    longCmdCall.apiData.containerType = request->containerType;
    longCmdCall.apiData.length = request->length;
    longCmdCall.apiData.transactionId = request->transactionId;
    memcpy(&longCmdCall.apiData.data, request->data, longCmdCall.apiData.length);

    longCmdCall.length = DS485N_TLV_HDR_LEN +
        DS485N_PKG_LEN(&longCmdCall.apiData);
    longCmdCall.command = DS485N_CMD_LONG_TX_PACKET;

    return sockwrite(cl, (uint8_t *) &longCmdCall, longCmdCall.length);
}

static int ds485_socket_client_ping(ds485_client_t *cl)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_PING;
    cmdCall.length = DS485N_TLV_HDR_LEN;

    return sockwrite(cl, (uint8_t *) &cmdCall, cmdCall.length);
}

static int ds485_reset_socket(ds485_client_t *cl)
{
    if (cl->sockfd < 0) {
        dlog(LOG_DEBUG, cl, "reset socket but handle is invalid\n");
        return 0;
    }

    dlog(LOG_DEBUG, cl, "reset socket\n");
    shutdown(cl->sockfd, SHUT_RDWR);
    close(cl->sockfd);
    cl->sockfd = -1;
    set_link_state(cl, false);
    cl->wait_pong = false;
    return 0;
}

/************************ SOCKET CONNECTION *******************************/

static int sockwrite(ds485_client_t *cl, uint8_t* data, size_t dlen)
{
    int rval;
    if (cl->sockfd < 0) {
        dlog(LOG_NOTICE, cl, "write: invalid socket\n");
        return ERROR_SOCKET_FAILED;
    }

    rval = socket_send(cl->sockfd, data, dlen);
    if (rval) {
        dlog(LOG_NOTICE, cl, "write: %s(%d)\n", strerror(rval), rval);
        close(cl->sockfd);
        cl->sockfd = -1;
        return ERROR_SOCKET_FAILED;
    }

    return 0;
}

/******************************* RECEIVE THREAD *******************************/

static int socket_service_dispatch(ds485_client_t *cl, uint8_t* data, uint16_t pktlen)
{
    dS485dCommand* cmd;
    int rval = 0;

    cmd = (dS485dCommand *) data;

    switch (cmd->command) {

    case DS485D_CMD_API_REVISION:
        if (pktlen != DS485N_TLV_HDR_LEN + 1) {
            dlog(LOG_WARNING, cl, "invalid DS485D_CMD_API_REVISION\n");
            /* strict what you emit, but tolerant what you accept */
        }
        dlog(LOG_NOTICE, cl, "Server API Revision: %d\n", cmd->data[0]);
        break;

    case DS485D_CMD_PING:
        /* allow every size */
        dlog(LOG_DEBUG, cl, "received pong\n");
        cl->wait_pong = false;
        break;

    case DS485D_CMD_TX_PACKET: {
        /* TODO is this a valid client command? */
        ds485_container_t* pData = (ds485_container_t *) cmd->data;
        if (pktlen != (DS485N_TLV_HDR_LEN + DS485N_CONT_LEN(pData))) {
            dlog(LOG_WARNING, cl, "short DS485D_CMD_TX_PACKET pktlen %d cont_len %d\n",
                pktlen, pData->length);
            /* strict what you emit, but tolerant what you accept */
        }
        break;
    }

    case DS485N_CMD_LONG_TX_PACKET: {
        /* TODO is this a valid client command? */
        ds485n_packet_t* pData = (ds485n_packet_t *) cmd->data;

        if (pktlen != (DS485N_TLV_HDR_LEN + DS485N_PKG_LEN(pData))) {
            dlog(LOG_WARNING, cl, "short DS485N_CMD_LONG_TX_PACKET pktlen %d cont_len %d\n",
                 pktlen, DS485N_PKG_LEN(pData));
            /* strict what you emit, but tolerant what you accept */
        }
        break;
    }

    case DS485D_CMD_RX_PACKET: {
        ds485_container_t container;
        ds485_container_t* pData = (ds485_container_t *) cmd->data;
        if (pktlen != (DS485N_TLV_HDR_LEN + DS485N_CONT_LEN(pData))) {
            dlog(LOG_WARNING, cl,
                 "wrapped container size mismatch: container: %d wrapper: %d\n",
                 DS485N_CONT_LEN(pData), pktlen);
            /* strict what you emit, but tolerant what you accept */
        }

        {
            /* copy data to fix alignment, just in case */
            char *tmp = (char *)&container;
            int plen = DS485N_CONT_LEN(pData);
            memcpy(tmp, pData, plen);
            memset(tmp + plen, 0, sizeof(container) - plen);
        }
        rval = do_container_received(cl, &container);
        break;
    }

    case DS485N_CMD_LONG_RX_PACKET: {
        ds485n_packet_t packet;
        ds485n_packet_t* pData = (ds485n_packet_t *) cmd->data;
        if (pktlen != DS485N_TLV_HDR_LEN + DS485N_PKG_LEN(pData)) {
            dlog(LOG_WARNING, cl,
                 "wrapped packet size mismatch: packet: %d wrapper %d\n",
                  DS485N_PKG_LEN(pData), pktlen);
            /* strict what you emit, but tolerant what you accept */
        }

        {
            /* copy data to fix alignment, just in case */
            int plen = DS485N_PKG_LEN(pData);
            char *tmp = (char *)&packet;
            memcpy(tmp, pData, plen);
            memset(tmp + plen, 0, sizeof(packet) - plen);
        }
        rval = do_long_container_received(cl, &packet);
        break;
    }

    case DS485D_CMD_QUERY_BUSSTATE:
        if (pktlen != DS485N_TLV_HDR_LEN + 1) {
            dlog(LOG_WARNING, cl, "invalid DS485D_CMD_QUERY_BUSSTATE\n");
            /* strict what you emit, but tolerant what you accept */
        }

        rval = do_link_state_reply_received(cl, v1_link_state(cmd->data[0]));
        break;

    case DS485D_CMD_QUERY_DEVICES:
        if ((pktlen - DS485N_TLV_HDR_LEN) % sizeof(dsuid_t) != 0) {
            dlog(LOG_WARNING, cl, "invalid device list len: %d\n", pktlen - DS485N_TLV_HDR_LEN);
            /* corrupt dsuid array, drop packet */
            break;
        }

        /* packet size limits maximum device list, no further check */
        rval = do_device_list_received(cl, cmd->data, pktlen - DS485N_TLV_HDR_LEN);
        break;

    case DS485D_CMD_BUS_CHANGE: {
        ds485_container_t* pData = (ds485_container_t *) cmd->data;
        if (pktlen != (DS485N_TLV_HDR_LEN + DS485N_CONT_LEN(pData))) {
            dlog(LOG_WARNING, cl, "short DS485D_CMD_BUS_CHANGE\n");
            /* strict what you emit, but tolerant what you accept */
        }

        if (pData->length > DS485_MAX_PAYLOAD_SIZE) {
            dlog(LOG_WARNING, cl, "dropping oversize bus_change: %d\n",
                 pData->length);
            break;
        }

        /* see DSM_LEFT */
        rval = do_bus_change_received(cl, pData->sourceId, pData->data[0]);
        break;
    }

    case DS485D_CMD_BUS_STATE: {
        ds485_container_t* pData = (ds485_container_t *) cmd->data;
        if (pktlen != (DS485N_TLV_HDR_LEN + DS485N_CONT_LEN(pData))) {
            dlog(LOG_WARNING, cl, "short DS485D_CMD_BUS_STATE\n");
            /* strict what you emit, but tolerant what you accept */
        }

        if (pData->length > DS485_MAX_PAYLOAD_SIZE) {
            dlog(LOG_WARNING, cl, "dropping oversize bus_state container: %d\n",
                 pData->length);
            break;
        }

        set_link_state(cl, v1_link_state(pData->data[0]));
        rval = 0;
        break;
    }

    case DS485D_CMD_QUERY_DSUID:
        if (pktlen != DS485N_TLV_HDR_LEN + sizeof(dsuid_t)) {
            dlog(LOG_WARNING, cl, "invalid DS485D_CMD_QUERY_DSUID\n");
            /* strict what you emit, but tolerant what you accept */
        }

        rval = do_dsid_query_received(cl, cmd->data);
        break;

    case DS485D_CMD_RESTART_BUS:
        if (pktlen != DS485N_TLV_HDR_LEN) {
            dlog(LOG_WARNING, cl, "invalid DS485D_CMD_RESTART_BUS\n");
            /* strict what you emit, but tolerant what you accept */
        }

        rval = do_restart_bus_received(cl);
        break;

    case DS485D_CMD_REGISTER:
    case DS485D_CMD_UNREGISTER:
    case DS485D_CMD_PACKETFILTER:
    case DS485D_CMD_QUERY_ADDRESS:
        dlog(LOG_NOTICE, cl, "invalid cmd:%s len:%d\n",
             cmd2text(cmd->command), pktlen);
        break;

    case DS485D_CMD_RAWMODE:
    case DS485D_CMD_DEVICE_PACKET:
    case DS485N_CMD_LONG_DEVICE_PACKET:
        dlog(LOG_NOTICE, cl, "deprecated cmd:%s len:%d\n",
             cmd2text(cmd->command), pktlen);
        break;

    default:
        /* invalid or protocol extension */
        dlog(LOG_NOTICE, cl, "unknown cmd:0x%02x len: %d\n", pktlen);
        break;
    }

    return rval;
}

static int client_connect(ds485_client_t *cl)
{
    int rval;

    cl->sockfd = socket_connect(cl->dName, cl->dPort, cl->dProtocol,
                                &cl->shutdown);
    if (cl->sockfd < 0) {
        set_link_state(cl, false);
        return ERROR_INVALID_FD;
    }

    /* valid local port number */
    update_pretty_name(cl);

    /* Set connection options */
    dS485dCommand cmd;
    cmd.command = DS485D_CMD_PACKETFILTER;
    cmd.length = DS485N_TLV_HDR_LEN + 1;
    cmd.data[0] = (cl->flags & PROMISCUOUS_MODE);
    rval = sockwrite(cl, (uint8_t *) &cmd, cmd.length);

    /* query bus state */
    cmd.command = DS485D_CMD_BUS_STATE;
    cmd.length = DS485N_TLV_HDR_LEN;
    rval = sockwrite(cl, (uint8_t *) &cmd, cmd.length);

    return rval;
}

static void* socket_thread(void* arg)
{
    ds485_client_t *cl = arg;
    int rval, idleTimer, unknown_cmd_count, cmd_ct;
    unsigned rxLen;
    struct pollfd fds[1];

#define RXBUFLEN 8500
    uint8_t* rxBuffer = (uint8_t *) malloc(RXBUFLEN);
    pthread_cleanup_push(free, rxBuffer);

    /* initialization before pthread_cleanup_push, produces warnings */
    idleTimer = unknown_cmd_count = 0;

    while (!cl->shutdown) {

        /* connect */
        if (cl->sockfd < 0) {

            rval = client_connect(cl);
            if (rval < 0) {
                dlog(LOG_NOTICE, cl, "reconnect failed\n");
                sleep(10);
                continue;
            }

            dlog(LOG_NOTICE, cl, "connected\n");
        }

        do_client_connect_success(cl);

        memset(&fds, 0, sizeof(fds));
        fds[0].fd = cl->sockfd;
        fds[0].events = POLLIN;

        /* as long as the socket is connected */
        cmd_ct = rxLen = 0;
        while (cl->sockfd > 0 && !cl->shutdown) {
            errno = 0;
            rval = poll(fds, 1, 1000);
            if (cl->shutdown) {
                break;
            }
            if (rval < 0) {
                if (errno == EINTR || errno == EINPROGRESS)
                    continue;
                dlog(LOG_NOTICE, cl, "client socket select error %d/%s\n",
                     errno, strerror(errno));
                continue;
            } else if (rval == 0) {
                if (idleTimer++ > 60) { /* 1 min */
                    idleTimer = 0;
                    if (cl->wait_pong) {
                        dlog(LOG_NOTICE, cl,
                             "keep alive timed out, restart connection..  processed %d cmd(s) in session\n",
                             cmd_ct);
                        ds485_reset_socket(cl);
                        break;
                    }

                    dlog(LOG_DEBUG, cl, "sending ping\n");
                    if (ds485_socket_client_ping(cl) < 0) {
                        ds485_reset_socket(cl);
                        break;
                    }
                    cl->wait_pong = true;
                }
            } else {
                int loop_count;

                if (fds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) {
                    dlog(LOG_WARNING, cl, "poll fd failed 0x%08x\n", fds[0].revents);
                    ds485_reset_socket(cl);
                    break;

                } else if (fds[0].revents & POLLIN) {
                    uint8_t* p;
                    rval = read(cl->sockfd, rxBuffer + rxLen, RXBUFLEN - rxLen);
                    if (rval < 0) {
                        dlog(LOG_NOTICE, cl, "client socket read error %s(%d)\n",
                             strerror(rval), rval);
                        continue;
                    }
                    else if (rval == 0) {
                        dlog(LOG_NOTICE, cl, "client socket closed by server\n");
                        ds485_reset_socket(cl);
                        break;
                    }

                    rxLen += rval;

                    p = rxBuffer;
                    loop_count = 0;
                    while (rxLen >= DS485N_TLV_HDR_LEN) {
                        unsigned pkt_len = p[2] << 8 | p[1];
                        int opcode = p[0];

                        /* TODO length check only for containers */
                        if (pkt_len < DS485N_TLV_HDR_LEN ||
                            ((opcode != DS485D_CMD_QUERY_DEVICES) &&
                             (pkt_len > (int)sizeof(dS485CmdLongApidata)))) {
                            dlog(LOG_ERR, cl,
                                 "Invalid %s message length %d, aborting\n",
                                 cmd2text(p[0]), pkt_len);
                            ds485_reset_socket(cl);
                            break;
                        }

                        if (loop_count++ > 71) {
                            dlog(LOG_ERR, cl,
                                 "too many attempts to read message, starting over\n");
                            ds485_reset_socket(cl);
                            break;
                        }

                        /* not enough data */
                        if (pkt_len > rxLen) {
                            break;
                        }

                        if (!strcmp(cmd2text(p[0]), "UNKNOWN")) {
                            unknown_cmd_count++;
                        } else {
                            unknown_cmd_count = 0;
                        }

                        if (unknown_cmd_count == 1) {
                            dlog(LOG_INFO, cl,
                                 "unknown cmd received prev cmd: %s len %d\n",
                                 cmd2text(cl->last_msg_opcode), cl->last_msg_len);
                        } else if (unknown_cmd_count > 100) {
                            dlog(LOG_ERR, cl,
                                 "too many unknown commands, starting over\n");
                            ds485_reset_socket(cl);
                            break;
                        }

                        cl->last_msg_opcode= p[0];
                        cl->last_msg_len = pkt_len;
                        cmd_ct++;

                        /* Process packet data ... */
                        rval = socket_service_dispatch(cl, p, pkt_len);
                        if (rval != 0) {
                            dlog(LOG_WARNING, cl,
                                 "socket_service_dispatch failed, aborting\n");
                            ds485_reset_socket(cl);
                            break;
                        }

                        p += pkt_len;
                        rxLen -= pkt_len;
                    }
                    if (p != rxBuffer && rxLen > 0) {
                        memmove(rxBuffer, p, rxLen);
                    }

                }
            }
        }

        if (!cl->shutdown) {
            // ratelimit retries
            sleep(5);
        }
    }

    ds485_reset_socket(cl);
    pthread_cleanup_pop(1);
    dlog(LOG_NOTICE, cl, "thread terminates\n");
    return NULL;
}
